"""Common CLI options used by Clinica pipelines."""

import click

from clinica.utils.pet import LIST_SUVR_REFERENCE_REGIONS

from .option_group import option

acq_label = option(
    "-acq",
    "--acq_label",
    help="Name of the label given to the PET acquisition, specifying the tracer used (acq-<acq_label>).",
)

dartel_tissues = option(
    "-dt",
    "--dartel_tissues",
    multiple=True,
    type=click.IntRange(1, 6),
    default=(1, 2, 3),
    show_default=True,
    help="Tissues to use for DARTEL template calculation.",
)

dont_save_warped_unmodulated = option(
    "-dswu",
    "--dont_save_warped_unmodulated",
    is_flag=True,
    help="Do not save warped unmodulated images for tissues specified in --tissue_classes flag.",
)

low_bval = option(
    "--low_bval",
    default=5,
    show_default=True,
    help="Define the b0 volumes as all volume bval <= low_bval.",
)

modulate = option(
    "--modulate/--no_modulate",
    "modulate",
    default=True,
    show_default=True,
    help="Modulate output images, no modulation preserves concentrations.",
)

n_procs = option(
    "-np",
    "--n_procs",
    type=int,
    help="Number of cores used to run in parallel.",
)

overwrite_outputs = option(
    "-overwrite",
    "--overwrite_outputs",
    is_flag=True,
    help="Force overwrite of output files in CAPS folder.",
)

pvc_psf_tsv = option(
    "-psf",
    "--pvc_psf_tsv",
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
    help=(
        "TSV file containing for each PET image its point spread function (PSF) measured "
        "in mm at x, y & z coordinates. Columns must contain: "
        "participant_id, session_id, acq_label, psf_x, psf_y and psf_z."
    ),
)

save_warped_modulated = option(
    "-swm",
    "--save_warped_modulated",
    is_flag=True,
    help="Save warped modulated images for tissues specified in --tissue_classes flag.",
)

smooth = option(
    "-s",
    "--smooth",
    multiple=True,
    default=(8,),
    show_default=True,
    help="Specify the different isomorphic FWHM in millimeters to smooth the image.",
)

subjects_sessions_tsv = option(
    "-tsv",
    "--subjects_sessions_tsv",
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
    help="TSV file containing a list of subjects with their sessions.",
)

suvr_reference_region = option(
    "-suvr",
    "--suvr_reference_region",
    type=click.Choice(LIST_SUVR_REFERENCE_REGIONS),
    help=(
        "Intensity normalization using the average PET uptake in reference regions "
        "resulting in a standardized uptake value ratio (SUVR) map. It can be "
        "cerebellumPons (used for amyloid tracers) or pons (used for 18F-FDG tracers)."
    ),
)

tissues = option(
    "-t",
    "--tissues",
    multiple=True,
    type=click.IntRange(1, 6),
    default=(1, 2, 3),
    show_default=True,
    help="Tissues to create flow fields to DARTEL template.",
)

tissue_classes = option(
    "-tc",
    "--tissue_classes",
    multiple=True,
    type=click.IntRange(1, 6),
    default=(1, 2, 3),
    show_default=True,
    help=(
        "Tissue classes (1: gray matter (GM), 2: white matter (WM), 3: cerebrospinal fluid (CSF), 4: bone, "
        "5: soft-tissue, 6: background) to save."
    ),
)

tissue_probability_maps = option(
    "-tpm",
    "--tissue_probability_maps",
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
    help="Tissue probability maps to use for segmentation.",
)

use_pvc_data = option(
    "-pvc",
    "--use_pvc_data",
    is_flag=True,
    help="Use PET data with partial value correction.",
)

voxel_size = option(
    "-vs",
    "--voxel_size",
    type=click.Tuple([float, float, float]),
    default=(1.5, 1.5, 1.5),
    show_default=True,
    help="Specifying the voxel sizes of the output image.",
)

working_directory = option(
    "-wd",
    "--working_directory",
    type=click.Path(file_okay=False, writable=True, resolve_path=True),
    help="Temporary directory to store pipelines intermediate results.",
)

yes = option(
    "-y",
    "--yes",
    is_flag=True,
    help="Execute the pipeline even if input images are not centered without asking for more user input.",
)



# coding: utf8

"""Define the command line parser for each utility (pipeline, iotools, etc.)."""

import abc
from argparse import ArgumentParser
from os import getcwd
from os.path import expanduser, join

PIPELINE_CATEGORIES = {
    "CLINICA_COMPULSORY": "Clinica mandatory arguments",
    "OPTIONAL": "Pipeline options",
    "CLINICA_OPTIONAL": "Clinica standard options",
    "ADVANCED": "Pipelines advanced options",
    "IOTOOLS_OPTIONS": "Optional arguments",
}


class CmdParser:
    """Abstract class to extend in order to create your command line parser.

    For pipelines, please use the 'clinica generate template' command.
    For converters, see clinica/iotools/converters/*/*_cli.py
    For iotools, see clinica/iotools/utils/data_handling_cli.py
    """

    __metaclass__ = abc.ABCMeta

    def __init__(self):
        self.reset()
        self.build()

    def build(self):
        self.define_name()
        self.define_description()
        self.set_content()
        self.define_options()

    def reset(self):
        self._args = ArgumentParser()
        self._name = None

    def set_content(self):
        self._args._positionals.title = "Mandatory arguments"
        self._args._optionals.title = "Optional arguments"
        if not self._description:
            self._description = self._name
            self._args.description = (
                "If you are not familiar with Clinica, see:\n"
                "https://aramislab.paris.inria.fr/clinica/docs/public/latest/InteractingWithClinica/"
            )
        else:
            self._args.description = (
                f"{self._description}\n\nIf you are not familiar with Clinica, see:\n"
                "https://aramislab.paris.inria.fr/clinica/docs/public/latest/InteractingWithClinica/"
            )

    @property
    def options(self):
        return self._args

    @options.setter
    def options(self, x):
        self._args = x

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, x):
        self._name = x

    @property
    def description(self):
        return self._description

    @description.setter
    def description(self, x):
        self._description = x

    @abc.abstractmethod
    def define_name(self):
        pass

    def define_description(self):
        self._description = None

    @abc.abstractmethod
    def define_options(self):
        pass

    def add_clinica_standard_arguments(
        self,
        add_tsv_flag=True,
        add_wd_flag=True,
        add_nprocs_flag=True,
        add_overwrite_flag=False,
        add_yes_flag=False,
    ):
        clinica_standard_options = self._args.add_argument_group(
            PIPELINE_CATEGORIES["CLINICA_OPTIONAL"]
        )
        if add_tsv_flag:
            clinica_standard_options.add_argument(
                "-tsv",
                "--subjects_sessions_tsv",
                help="TSV file containing a list of subjects with their sessions.",
            )
        if add_wd_flag:
            clinica_standard_options.add_argument(
                "-wd",
                "--working_directory",
                help="Temporary directory to store pipelines intermediate results.",
            )
        if add_nprocs_flag:
            clinica_standard_options.add_argument(
                "-np",
                "--n_procs",
                metavar="N",
                type=int,
                help="Number of cores used to run in parallel.",
            )
        if add_overwrite_flag:
            clinica_standard_options.add_argument(
                "-overwrite",
                "--overwrite_outputs",
                action="store_true",
                default=False,
                help="Force overwrite of output files in CAPS folder.",
            )

        if add_yes_flag:
            clinica_standard_options.add_argument(
                "-y",
                "--yes",
                action="store_true",
                default=False,
                help="Execute the pipeline even if input images are not centered without "
                "asking for more user input.",
            )

        return clinica_standard_options

    @abc.abstractmethod
    def run_command(self, args):
        pass

    @staticmethod
    def list_to_string(list):
        """Convert list (e.g. [8, 8, 8]) to string (e.g. '8 8 8')."""
        string_without_commas_and_brackets = " ".join(str(item) for item in list)
        return string_without_commas_and_brackets

    @staticmethod
    def absolute_path(arg):
        if not arg:
            return None
        elif arg[:1] == "~":
            return expanduser(arg)
        elif arg[:2] == "./":
            return join(getcwd(), arg[2:])
        else:
            return join(getcwd(), arg)


def init_cmdparser_objects(root_parser, parser, objects):
    """Init all derived CmdParser instances with specific data.

    Args:
        root_parser: The root parser
        parser: The ArgParser node (e.g. 'run' or 'convert')
        objects: All CmdParser instances of this file
    """

    def silent_help():
        pass

    def error_message(p):
        def error(x):
            p.print_help()
            root_parser.print_help = silent_help
            exit(-1)

        return error

    def init(x):
        import argparse

        x.options = parser.add_parser(
            x.name,
            add_help=False,
            help=x.description,
            formatter_class=argparse.RawDescriptionHelpFormatter,
        )
        x.options.error = error_message(x.options)
        x.options.set_defaults(func=x.run_command)
        x.build()

    for x in objects:
        try:
            init(x)
        except BaseException:
            pass



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import os.path as op

from nipype.interfaces.base import (
    CommandLine,
    CommandLineInputSpec,
    File,
    InputMultiPath,
    TraitedSpec,
    isdefined,
    traits,
)
from nipype.interfaces.mrtrix3.base import MRTrix3Base, MRTrix3BaseInputSpec


class TensorMetricsInputSpec(CommandLineInputSpec):
    in_file = File(
        exists=True, argstr="%s", mandatory=True, position=-1, desc="input DTI image"
    )

    out_fa = File(argstr="-fa %s", desc="output FA file")
    out_adc = File(argstr="-adc %s", desc="output ADC file")
    out_ad = File(argstr="-ad %s", desc="output AD file")
    out_rd = File(argstr="-rd %s", desc="output RD file")
    out_evec = File(argstr="-vector %s", desc="output selected eigenvector(s) file")
    out_eval = File(argstr="-value %s", desc="output selected eigenvalue(s) file")
    component = traits.List(
        [1],
        usedefault=True,
        argstr="-num %s",
        sep=",",
        desc=(
            "specify the desired eigenvalue/eigenvector(s). Note that "
            "several eigenvalues can be specified as a number sequence"
        ),
    )
    in_mask = File(
        exists=True,
        argstr="-mask %s",
        desc=(
            "only perform computation within the specified binary" " brain mask image"
        ),
    )
    modulate = traits.Enum(
        "FA",
        "none",
        "eval",
        argstr="-modulate %s",
        desc=("how to modulate the magnitude of the" " eigenvectors"),
    )


class TensorMetricsOutputSpec(TraitedSpec):
    out_fa = File(desc="output FA file")
    out_adc = File(desc="output ADC file")
    out_ad = File(desc="output AD file")
    out_rd = File(desc="output RD file")
    out_evec = File(desc="output selected eigenvector(s) file")
    out_eval = File(desc="output selected eigenvalue(s) file")


class TensorMetrics(CommandLine):
    """
    Compute metrics from tensors


    Example
    -------

    >>> import nipype.interfaces.mrtrix3 as mrt
    >>> comp = mrt.TensorMetrics()
    >>> comp.inputs.in_file = 'dti.mif'
    >>> comp.inputs.out_fa = 'fa.mif'
    >>> comp.cmdline                               # doctest: +ELLIPSIS
    'tensor2metric -num 1 -fa fa.mif dti.mif'
    >>> comp.run()                                 # doctest: +SKIP
    """

    _cmd = "tensor2metric"
    input_spec = TensorMetricsInputSpec
    output_spec = TensorMetricsOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()

        for k in list(outputs.keys()):
            if isdefined(getattr(self.inputs, k)):
                outputs[k] = op.abspath(getattr(self.inputs, k))

        return outputs



# coding: utf8


def susceptibility_distortion_correction_using_t1(
        name='susceptibility_distortion_correction_using_t1'):
    """
    Susceptibility distortion correction using the T1w image.

    This workflow allows to correct for echo-planar induced susceptibility
    artifacts without fieldmap (e.g. ADNI Database) by elastically register
    DWIs to their respective baseline T1-weighted structural scans using an
    inverse consistent registration algorithm with a mutual information cost
    function (SyN algorithm).

    Args:
        name (Optional[str]): Name of the workflow.

    Inputnode:
        in_t1 (str): T1w image.
        in_dwi (str): DWI dataset

    Outputnode:
        out_dwi (str): Corrected DWI dataset
        out_warp (str): Out warp allowing DWI to T1 registration and
            susceptibilty induced artifacts correction
        out_b0_to_t1_rigid_body_matrix (str): B0 to T1 image FLIRT rigid body
            FSL coregistration matrix
        out_t1_to_b0_rigid_body_matrix (str): T1 to B0 image FLIRT rigid body
            FSL coregistration matrix
        out_t1_coregistered_to_b0 (str): T1 image rigid body coregistered to
            the B0 image
        out_b0_to_t1_syn_deformation_field (str): B0 to T1 image ANTs SyN
            ITK warp
        out_b0_to_t1_affine_matrix (str): B0 to T1 image ANTs affine ITK
            coregistration matrix

    References:
      .. Nir et al. (Neurobiology of Aging 2015): Connectivity network measures
        predict volumetric atrophy in mild cognitive impairment

      .. Leow et al. (IEEE Trans Med Imaging 2007): Statistical Properties of
        Jacobian Maps and the Realization of Unbiased Large Deformation
        Nonlinear Image Registration


    Returns:
        The workflow

    Example:
        >>> epi = susceptibility_distortion_correction_using_t1()
        >>> epi.inputs.inputnode.in_dwi = 'dwi.nii'
        >>> epi.inputs.inputnode.in_t1 = 'T1w.nii'
        >>> epi.run() # doctest: +SKIP
    """
    import nipype
    import nipype.interfaces.fsl as fsl
    import nipype.interfaces.utility as niu
    import nipype.pipeline.engine as pe

    import clinica.pipelines.dwi_preprocessing_using_t1.dwi_preprocessing_using_t1_utils as utils

    def expend_matrix_list(in_matrix, in_bvec):
        import numpy as np

        bvecs = np.loadtxt(in_bvec).T
        out_matrix_list = [in_matrix]

        out_matrix_list = out_matrix_list * len(bvecs)

        return out_matrix_list

    def rotate_bvecs(in_bvec, in_matrix):
        """
        Rotates the input bvec file accordingly with a list of matrices.
        .. note:: the input affine matrix transforms points in the destination
          image to their corresponding coordinates in the original image.
          Therefore, this matrix should be inverted first, as we want to know
          the target position of :math:`\\vec{r}`.
        """
        import os

        import numpy as np

        name, fext = os.path.splitext(os.path.basename(in_bvec))
        if fext == '.gz':
            name, _ = os.path.splitext(name)
        out_file = os.path.abspath('%s_rotated.bvec' % name)
        bvecs = np.loadtxt(
            in_bvec).T  # Warning, bvecs.txt are not in the good configuration, need to put '.T'
        new_bvecs = []

        if len(bvecs) != len(in_matrix):
            raise RuntimeError(('Number of b-vectors (%d) and rotation '
                                'matrices (%d) should match.') %
                               (len(bvecs), len(in_matrix)))

        for bvec, mat in zip(bvecs, in_matrix):
            if np.all(bvec == 0.0):
                new_bvecs.append(bvec)
            else:
                invrot = np.linalg.inv(np.loadtxt(mat))[:3, :3]
                newbvec = invrot.dot(bvec)
                new_bvecs.append((newbvec / np.linalg.norm(newbvec)))

        np.savetxt(out_file, np.array(new_bvecs).T, fmt='%0.15f')
        return out_file

    inputnode = pe.Node(
        niu.IdentityInterface(fields=['in_t1', 'in_dwi', 'in_bvec']),
        name='inputnode')

    split = pe.Node(fsl.Split(dimension='t'), name='SplitDWIs')
    pick_ref = pe.Node(niu.Select(), name='Pick_b0')
    pick_ref.inputs.index = [0]

    flirt_b0_to_t1 = pe.Node(interface=fsl.FLIRT(dof=6),
                             name='flirt_b0_to_t1')
    flirt_b0_to_t1.inputs.interp = "spline"
    flirt_b0_to_t1.inputs.cost = 'normmi'
    flirt_b0_to_t1.inputs.cost_func = 'normmi'

    if nipype.__version__.split('.') < ['0', '13', '0']:
        apply_xfm = pe.Node(interface=fsl.ApplyXfm(),
                            name='apply_xfm')
    else:
        apply_xfm = pe.Node(interface=fsl.ApplyXFM(),
                            name='apply_xfm')
    apply_xfm.inputs.apply_xfm = True

    expend_matrix = pe.Node(
        interface=niu.Function(input_names=['in_matrix', 'in_bvec'],
                               output_names=['out_matrix_list'],
                               function=expend_matrix_list),
        name='expend_matrix')

    rot_bvec = pe.Node(niu.Function(input_names=['in_matrix', 'in_bvec'],
                                    output_names=['out_file'],
                                    function=rotate_bvecs),
                       name='Rotate_Bvec')

    ants_registration_syn_quick = pe.Node(interface=niu.Function(
        input_names=['fix_image', 'moving_image'],
        output_names=['image_warped', 'affine_matrix',
                      'warp', 'inverse_warped', 'inverse_warp'],
        function=utils.ants_registration_syn_quick),
        name='ants_registration_syn_quick')

    merge_transform = pe.Node(niu.Merge(2), name='MergeTransforms')

    combine_warp = pe.Node(interface=niu.Function(
        input_names=['in_file', 'transforms_list', 'reference'],
        output_names=['out_warp'],
        function=utils.ants_combine_transform), name='combine_warp')

    coeffs = pe.Node(fsl.WarpUtils(out_format='spline'), name='CoeffComp')

    fsl_transf = pe.Node(fsl.WarpUtils(out_format='field'),
                         name='fsl_transf')

    warp_epi = pe.Node(fsl.ConvertWarp(), name='warp_epi')

    apply_warp = pe.MapNode(interface=fsl.ApplyWarp(),
                            iterfield=['in_file'], name='apply_warp')
    apply_warp.inputs.interp = 'spline'

    thres = pe.MapNode(fsl.Threshold(thresh=0.0), iterfield=['in_file'],
                       name='RemoveNegative')

    merge = pe.Node(fsl.Merge(dimension='t'), name='MergeDWIs')

    outputnode = pe.Node(niu.IdentityInterface(
        fields=['dwi_to_t1_coregistration_matrix',
                'itk_dwi_t1_coregistration_matrix',
                'epi_correction_deformation_field',
                'epi_correction_affine_transform',
                'merge_epi_transform', 'out_dwi', 'out_warp',
                'out_bvec']), name='outputnode')

    wf = pe.Workflow(name=name)

    wf.connect([
        (inputnode, split, [('in_dwi', 'in_file')]),  # noqa

        (split, pick_ref, [('out_files', 'inlist')]),  # noqa

        (pick_ref, flirt_b0_to_t1, [('out', 'in_file')]),  # noqa
        (inputnode, flirt_b0_to_t1, [('in_t1', 'reference')]),  # noqa

        (flirt_b0_to_t1, expend_matrix, [('out_matrix_file', 'in_matrix')]),
        # noqa
        (inputnode, expend_matrix, [('in_bvec', 'in_bvec')]),  # noqa

        (inputnode, rot_bvec, [('in_bvec', 'in_bvec')]),  # noqa
        (expend_matrix, rot_bvec, [('out_matrix_list', 'in_matrix')]),  # noqa

        (inputnode, ants_registration_syn_quick, [('in_t1', 'fix_image')]),
        # noqa
        (flirt_b0_to_t1, ants_registration_syn_quick,
         [('out_file', 'moving_image')]),  # noqa

        (ants_registration_syn_quick, merge_transform,
         [('affine_matrix', 'in2'),  # noqa
          ('warp', 'in1')]),  # noqa

        (flirt_b0_to_t1, combine_warp, [('out_file', 'in_file')]),  # noqa
        (merge_transform, combine_warp, [('out', 'transforms_list')]),  # noqa
        (inputnode, combine_warp, [('in_t1', 'reference')]),  # noqa

        (inputnode, coeffs, [('in_t1', 'reference')]),  # noqa
        (combine_warp, coeffs, [('out_warp', 'in_file')]),  # noqa

        (coeffs, fsl_transf, [('out_file', 'in_file')]),  # noqa
        (inputnode, fsl_transf, [('in_t1', 'reference')]),  # noqa

        (inputnode, warp_epi, [('in_t1', 'reference')]),  # noqa
        (flirt_b0_to_t1, warp_epi, [('out_matrix_file', 'premat')]),  # noqa
        (fsl_transf, warp_epi, [('out_file', 'warp1')]),  # noqa

        (warp_epi, apply_warp, [('out_file', 'field_file')]),  # noqa
        (split, apply_warp, [('out_files', 'in_file')]),  # noqa
        (inputnode, apply_warp, [('in_t1', 'ref_file')]),  # noqa

        (apply_warp, thres, [('out_file', 'in_file')]),  # noqa

        (thres, merge, [('out_file', 'in_files')]),  # noqa
        # Outputnode
        (merge, outputnode, [('merged_file', 'out_dwi')]),  # noqa
        (flirt_b0_to_t1, outputnode,
         [('out_matrix_file', 'dwi_to_t1_coregistration_matrix')]),  # noqa
        (ants_registration_syn_quick, outputnode,
         [('warp', 'epi_correction_deformation_field'),  # noqa
          ('affine_matrix', 'epi_correction_affine_transform')]),  # noqa
        (warp_epi, outputnode, [('out_file', 'out_warp')]),  # noqa
        (rot_bvec, outputnode, [('out_file', 'out_bvec')]),  # noqa
    ])
    return wf


def apply_all_corrections_using_ants(name='UnwarpArtifacts'):
    """
    Combines two lists of linear transforms with the deformation field
    map obtained epi_correction by Ants.
    Additionally, computes the corresponding bspline coefficients and
    the map of determinants of the jacobian.
    """
    import nipype.interfaces.fsl as fsl
    import nipype.interfaces.utility as niu
    import nipype.pipeline.engine as pe

    inputnode = pe.Node(niu.IdentityInterface(
        fields=['in_sdc_syb', 'in_hmc', 'in_ecc', 'in_dwi', 'in_t1']),
        name='inputnode')

    outputnode = pe.Node(niu.IdentityInterface(
        fields=['out_file', 'out_warp', 'out_coeff', 'out_jacobian']),
        name='outputnode')

    split = pe.Node(fsl.Split(dimension='t'), name='SplitDWIs')

    concat_hmc_ecc = pe.MapNode(fsl.ConvertXFM(), name="concat_hmc_ecc",
                                iterfield=['in_file', 'in_file2'])
    concat_hmc_ecc.inputs.concat_xfm = True

    warps = pe.MapNode(fsl.ConvertWarp(), iterfield=['premat'],
                       name='ConvertWarp')

    unwarp = pe.MapNode(interface=fsl.ApplyWarp(),
                        iterfield=['in_file', 'field_file'],
                        name='unwarp_warp')
    unwarp.inputs.interp = 'spline'

    coeffs = pe.MapNode(fsl.WarpUtils(out_format='spline'),
                        iterfield=['in_file'], name='CoeffComp')
    jacobian = pe.MapNode(fsl.WarpUtils(write_jacobian=True),
                          iterfield=['in_file'], name='JacobianComp')
    jacmult = pe.MapNode(fsl.MultiImageMaths(op_string='-mul %s'),
                         iterfield=['in_file', 'operand_files'],
                         name='ModulateDWIs')

    thres = pe.MapNode(fsl.Threshold(thresh=0.0), iterfield=['in_file'],
                       name='RemoveNegative')
    merge = pe.Node(fsl.Merge(dimension='t'), name='MergeDWIs')

    wf = pe.Workflow(name=name)
    wf.connect([
        (inputnode, concat_hmc_ecc, [('in_ecc', 'in_file2')]),  # noqa
        (inputnode, concat_hmc_ecc, [('in_hmc', 'in_file')]),  # noqa

        (concat_hmc_ecc, warps, [('out_file', 'premat')]),  # noqa
        (inputnode, warps, [('in_sdc_syb', 'warp1')]),  # noqa
        (inputnode, warps, [('in_t1', 'reference')]),  # noqa
        (inputnode, split, [('in_dwi', 'in_file')]),  # noqa
        (warps, unwarp, [('out_file', 'field_file')]),  # noqa
        (split, unwarp, [('out_files', 'in_file')]),  # noqa
        (inputnode, unwarp, [('in_t1', 'ref_file')]),  # noqa
        (inputnode, coeffs, [('in_t1', 'reference')]),  # noqa
        (warps, coeffs, [('out_file', 'in_file')]),  # noqa
        (inputnode, jacobian, [('in_t1', 'reference')]),  # noqa
        (coeffs, jacobian, [('out_file', 'in_file')]),  # noqa
        (unwarp, jacmult, [('out_file', 'in_file')]),  # noqa
        (jacobian, jacmult, [('out_jacobian', 'operand_files')]),  # noqa
        (jacmult, thres, [('out_file', 'in_file')]),  # noqa
        (thres, merge, [('out_file', 'in_files')]),  # noqa
        (warps, outputnode, [('out_file', 'out_warp')]),  # noqa
        (coeffs, outputnode, [('out_file', 'out_coeff')]),  # noqa
        (jacobian, outputnode, [('out_jacobian', 'out_jacobian')]),  # noqa
        (merge, outputnode, [('merged_file', 'out_file')])  # noqa
    ])

    return wf



from typing import Optional

import click

from clinica.pipelines import cli_param

pipeline_name = "t1-freesurfer-longitudinal"


@click.command(name=pipeline_name)
@cli_param.argument.caps_directory
@cli_param.option_group.common_pipelines_options
@cli_param.option.subjects_sessions_tsv
@cli_param.option.working_directory
@cli_param.option.n_procs
@cli_param.option.overwrite_outputs
@click.pass_context
def cli(
    ctx: click.Context,
    caps_directory: str,
    subjects_sessions_tsv: Optional[str] = None,
    working_directory: Optional[str] = None,
    n_procs: Optional[int] = None,
    overwrite_outputs: bool = False,
) -> None:
    """Longitudinal pre-processing of T1w images with FreeSurfer.

    https://aramislab.paris.inria.fr/clinica/docs/public/latest/Pipelines/T1_FreeSurfer_Longitudinal/
    """
    import datetime
    import os

    from clinica.utils.longitudinal import get_participants_long_id
    from clinica.utils.participant import get_subject_session_list
    from clinica.utils.stream import cprint

    from . import t1_freesurfer_longitudinal_correction_cli, t1_freesurfer_template_cli
    from .longitudinal_utils import save_part_sess_long_ids_to_tsv

    cprint(
        "The t1-freesurfer-longitudinal pipeline is divided into 2 parts:\n"
        "\tt1-freesurfer-unbiased-template pipeline: Creation of unbiased template\n"
        "\tt1-freesurfer-longitudinal-correction pipeline: Longitudinal correction."
    )

    if not subjects_sessions_tsv:
        l_sess, l_part = get_subject_session_list(caps_directory, None, False, False)
        l_long = get_participants_long_id(l_part, l_sess)
        now = datetime.datetime.now().strftime("%H%M%S")
        subjects_sessions_tsv = now + "_participants.tsv"
        save_part_sess_long_ids_to_tsv(
            l_part, l_sess, l_long, os.getcwd(), subjects_sessions_tsv
        )

    cprint("Part 1/2: Running t1-freesurfer-unbiased-template pipeline.")
    ctx.invoke(
        t1_freesurfer_longitudinal_correction_cli.cli,
        caps_directory=caps_directory,
        subjects_sessions_tsv=subjects_sessions_tsv,
        working_directory=working_directory,
        n_procs=n_procs,
        overwrite_outputs=overwrite_outputs,
    )

    cprint("Part 2/2 Running t1-freesurfer-longitudinal-correction pipeline.")
    ctx.invoke(
        t1_freesurfer_longitudinal_correction_cli.cli,
        caps_directory=caps_directory,
        subjects_sessions_tsv=subjects_sessions_tsv,
        working_directory=working_directory,
        n_procs=n_procs,
        overwrite_outputs=overwrite_outputs,
    )


if __name__ == "__main__":
    cli()



# coding: utf-8

import clinica.pipelines.engine as cpe


class StatisticsVolumeCorrection(cpe.Pipeline):
    """StatisticsVolumeCorrection - Statistical correction of Statistical correction of StatisticsVolume pipeline.

    Returns:
        A clinica pipeline object containing the StatisticsVolumeCorrection pipeline.
    """

    def check_custom_dependencies(self):
        """Check dependencies that can not be listed in the `info.json` file."""

    def get_input_fields(self):
        """Specify the list of possible inputs of this pipeline.

        Returns:
            A list of (string) input fields name.
        """
        return ["t_map"]

    def get_output_fields(self):
        """Specify the list of possible outputs of this pipeline.

        Returns:
            A list of (string) output fields name.
        """
        return []

    def build_input_node(self):
        """Build and connect an input node to the pipeline."""
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        from clinica.utils.inputs import clinica_group_reader

        t_map = clinica_group_reader(
            self.caps_directory,
            {
                "pattern": self.parameters["t_map"] + "*",
                "description": "statistics t map",
                "needed_pipeline": "statistics-volume",
            },
        )

        read_parameters_node = npe.Node(
            name="LoadingCLIArguments",
            interface=nutil.IdentityInterface(
                fields=self.get_input_fields(), mandatory_inputs=True
            ),
        )
        read_parameters_node.inputs.t_map = t_map

        self.connect([(read_parameters_node, self.input_node, [("t_map", "t_map")])])

    def build_output_node(self):
        """Build and connect an output node to the pipeline."""

    def build_core_nodes(self):
        """Build and connect the core nodes of the pipeline."""
        from os.path import abspath, dirname, exists, join, pardir

        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe
        import numpy as np

        import clinica.pipelines.statistics_volume_correction.statistics_volume_correction_utils as utils
        from clinica.utils.inputs import RemoteFileStructure, fetch_file
        from clinica.utils.stream import cprint

        peak_correction_FWE = npe.Node(
            name="peak_correction_FWE",
            interface=nutil.Function(
                input_names=["t_map", "t_threshold"],
                output_names=["output"],
                function=utils.peak_correction,
            ),
        )
        peak_correction_FWE.inputs.t_threshold = self.parameters["FWEp"]

        peak_correction_FDR = peak_correction_FWE.clone(name="peak_correction_FDR")
        peak_correction_FDR.inputs.t_threshold = self.parameters["FDRp"]

        cluster_correction_FWE = npe.Node(
            name="cluster_correction_FWE",
            interface=nutil.Function(
                input_names=["t_map", "t_thresh", "c_thresh"],
                output_names=["output"],
                function=utils.cluster_correction,
            ),
        )
        cluster_correction_FWE.inputs.t_thresh = self.parameters["height_threshold"]
        cluster_correction_FWE.inputs.c_thresh = self.parameters["FWEc"]

        cluster_correction_FDR = cluster_correction_FWE.clone(
            name="cluster_correction_FDR"
        )
        cluster_correction_FDR.inputs.t_thresh = self.parameters["height_threshold"]
        cluster_correction_FDR.inputs.c_thresh = self.parameters["FDRc"]

        produce_fig_FWE_peak_correction = npe.Node(
            name="produce_figure_FWE_peak_correction",
            interface=nutil.Function(
                input_names=[
                    "nii_file",
                    "template",
                    "type_of_correction",
                    "t_thresh",
                    "c_thresh",
                    "n_cuts",
                ],
                output_names=["figs"],
                function=utils.produce_figures,
            ),
        )
        produce_fig_FWE_peak_correction.inputs.n_cuts = self.parameters["n_cuts"]

        root = dirname(abspath(join(abspath(__file__), pardir, pardir)))
        path_to_mask = join(root, "resources", "masks")
        url_aramis = "https://aramislab.paris.inria.fr/files/data/img_t1_linear/"
        FILE1 = RemoteFileStructure(
            filename="mni_icbm152_t1_tal_nlin_sym_09a.nii.gz",
            url=url_aramis,
            checksum="3b244ee7e287319d36a25263744c468ef0ab2fe5a94b15a2138844db73b49adf",
        )

        produce_fig_FWE_peak_correction.inputs.template = join(
            path_to_mask, FILE1.filename
        )
        if not (exists(produce_fig_FWE_peak_correction.inputs.template)):
            try:
                fetch_file(FILE1, path_to_mask)
            except IOError as err:
                cprint(
                    msg=f"Unable to download required template (mni_icbm152) for processing: {err}",
                    lvl="error",
                )

        produce_fig_FDR_peak_correction = produce_fig_FWE_peak_correction.clone(
            name="produce_figure_FDR_peak_correction"
        )
        produce_fig_FWE_cluster_correction = produce_fig_FWE_peak_correction.clone(
            name="produce_figure_FWE_cluster_correction"
        )
        produce_fig_FDR_cluster_correction = produce_fig_FWE_peak_correction.clone(
            name="produce_figure_FDR_cluster_correction"
        )

        # fmt: off
        produce_fig_FWE_peak_correction.inputs.type_of_correction = "FWE"
        produce_fig_FDR_peak_correction.inputs.type_of_correction = "FDR"
        produce_fig_FWE_cluster_correction.inputs.type_of_correction = "FWE"
        produce_fig_FDR_cluster_correction.inputs.type_of_correction = "FDR"

        produce_fig_FWE_peak_correction.inputs.t_thresh = self.parameters["FWEp"]
        produce_fig_FDR_peak_correction.inputs.t_thresh = self.parameters["FDRp"]
        produce_fig_FWE_cluster_correction.inputs.t_thresh = self.parameters["height_threshold"]
        produce_fig_FDR_cluster_correction.inputs.t_thresh = self.parameters["height_threshold"]

        produce_fig_FWE_peak_correction.inputs.c_thresh = np.nan
        produce_fig_FDR_peak_correction.inputs.c_thresh = np.nan
        produce_fig_FWE_cluster_correction.inputs.c_thresh = self.parameters["FWEc"]
        produce_fig_FDR_cluster_correction.inputs.c_thresh = self.parameters["FDRc"]
        # fmt: on

        save_fig_peak_correction_FWE = npe.Node(
            name="save_figure_peak_correction_FWE",
            interface=nutil.Function(
                input_names=["t_map", "figs", "name"],
                output_names=[],
                function=utils.generate_output,
            ),
        )
        save_fig_peak_correction_FWE.inputs.name = "FWEp"

        save_fig_peak_correction_FDR = save_fig_peak_correction_FWE.clone(
            name="save_fig_peak_correction_FDR"
        )
        save_fig_peak_correction_FDR.inputs.name = "FDRp"

        save_fig_cluster_correction_FWE = save_fig_peak_correction_FWE.clone(
            name="save_fig_cluster_correction_FWE"
        )
        save_fig_cluster_correction_FWE.inputs.name = "FWEc"

        save_fig_cluster_correction_FDR = save_fig_peak_correction_FWE.clone(
            name="save_fig_cluster_correction_FDR"
        )
        save_fig_cluster_correction_FDR.inputs.name = "FDRc"

        # Connection
        # ==========
        # fmt: off
        self.connect(
            [
                (self.input_node, peak_correction_FWE, [("t_map", "t_map")]),
                (self.input_node, peak_correction_FDR, [("t_map", "t_map")]),
                (self.input_node, cluster_correction_FWE, [("t_map", "t_map")]),
                (self.input_node, cluster_correction_FDR, [("t_map", "t_map")]),
                (peak_correction_FWE, produce_fig_FWE_peak_correction, [("output", "nii_file")]),
                (peak_correction_FDR, produce_fig_FDR_peak_correction, [("output", "nii_file")]),
                (cluster_correction_FWE, produce_fig_FWE_cluster_correction, [("output", "nii_file")]),
                (cluster_correction_FDR, produce_fig_FDR_cluster_correction, [("output", "nii_file")]),
                (produce_fig_FWE_peak_correction, save_fig_peak_correction_FWE, [("figs", "figs")]),
                (produce_fig_FDR_peak_correction, save_fig_peak_correction_FDR, [("figs", "figs")]),
                (produce_fig_FWE_cluster_correction, save_fig_cluster_correction_FWE, [("figs", "figs")]),
                (produce_fig_FDR_cluster_correction, save_fig_cluster_correction_FDR, [("figs", "figs")]),
                (self.input_node, save_fig_peak_correction_FWE, [("t_map", "t_map")]),
                (self.input_node, save_fig_peak_correction_FDR, [("t_map", "t_map")]),
                (self.input_node, save_fig_cluster_correction_FWE, [("t_map", "t_map")]),
                (self.input_node, save_fig_cluster_correction_FDR, [("t_map", "t_map")]),
            ]
        )
